from copy import deepcopy
from collections import defaultdict
from itertools import chain
import re


def get_units(ent_type, doc, doc_id=None, enrich_container=[]):
    relation_layer = "Scholarly"
    doc = deepcopy(doc)
    lines = generate_line_annos(doc["text"])
    doc["annotations"]["Line"] = lines
    units = list(get_ents(ent_type, doc["text"], doc["annotations"], doc_id))
    for unit in units:
        # reformat begin and end
        unit["begin"] = unit["begin_in_doc"]
        unit["end"] = unit["end_in_doc"]
        del unit["begin_in_doc"]
        del unit["end_in_doc"]
        container_dict = {}
        for container in unit["container"]:
            if container["type"] in enrich_container:
                begin = container["begin_in_doc"]
                end = container["end_in_doc"]
                container["text"] = doc["text"][begin:end]
            if container["type"] in container:
                raise Exception("More than one container of one type")
            container_dict[container["type"]] = container
        unit["container"] = container_dict
    add_relations(units, doc.get("relations", {}).get("Scholarly"))
    return units


def add_relations(units, relations):
    if not relations:
        return
    # match with annotations of units
    source_ent_relations = defaultdict(list)
    target_ent_relations = defaultdict(list)
    for relation_idx, relation in enumerate(relations):
        relation["idx"] = relation_idx
        source_ent_relations[relation["source_id"]].append(relation_idx)
        target_ent_relations[relation["target_id"]].append(relation_idx)

    for unit in units:
        ent_ids = {a.get("ref_id") for a in unit["annotations"] if "ref_id" in a}

        relations_idxs_source = set(
            chain(*[source_ent_relations[e_id] for e_id in ent_ids])
        )
        relations_idxs_target = set(
            chain(*[target_ent_relations[e_id] for e_id in ent_ids])
        )
        relations_idxs_inner = relations_idxs_source & relations_idxs_target
        units_relations = [
            deepcopy(relations[i])
            for i in list(relations_idxs_source | relations_idxs_target)
        ]
        for rel in units_relations:
            rel["source_in_unit"] = rel["idx"] in relations_idxs_source
            rel["target_in_unit"] = rel["idx"] in relations_idxs_target
            rel["inner"] = rel["idx"] in relations_idxs_inner
        unit["relations"] = units_relations
    # match with unit itself
    # @todo


def get_ents(ent_type, text, annos, doc_id=None):
    other_ents = [k for k in annos.keys() if k != ent_type]
    left_annos = deepcopy({k: v[::-1] for k, v in annos.items() if k != ent_type})
    for ent in annos.get(ent_type, []):
        ent_info = {k: v for k, v in ent.items()}
        ent_info["type"] = ent_type
        ent_info["text"] = text[ent["begin"] : ent["end"]]
        begin, end = ent["begin"], ent["end"]
        found_ents = update_left_annos(left_annos, begin, end)
        found_ents = list(found_ents)
        container_ents = [
            e for e, ent_flag in found_ents if ent_flag == "container_ent"
        ]
        # sort container by size (biggest (most genral) first)
        container_ents.sort(key=lambda x: -x["end_in_doc"] + x["begin_in_doc"])
        sub_ents = [e for e, ent_flag in found_ents if ent_flag == "sub_ent"]
        overlapping_ents = [
            e for e, ent_flag in found_ents if ent_flag == "overlapping_ent"
        ]
        ent_info["annotations"] = sub_ents
        ent_info["container"] = container_ents
        if overlapping_ents:
            ent_info["annotations_overlap"] = overlapping_ents
        ent_info["begin_in_doc"] = ent_info["begin"]
        del ent_info["begin"]
        ent_info["end_in_doc"] = ent_info["end"]
        del ent_info["end"]
        if doc_id is not None:
            ent_info["doc_id"] = doc_id
        yield ent_info


def update_left_annos(left_annos, begin, end):
    for key, values in left_annos.items():
        while True:
            if not values:
                break
            next_ent = values[-1]
            if next_ent["end"] <= begin:
                # span is before
                values.pop()
                continue
            elif next_ent["begin"] <= begin and next_ent["end"] >= end:
                # span is container span of this span
                next_ent["begin_in_doc"] = next_ent["begin"]
                next_ent["end_in_doc"] = next_ent["end"]
                next_ent["offset_to_container"] = begin - next_ent["begin_in_doc"]
                # next_ent["end_in_container"] = end - next_ent["begin_in_doc"]
                next_ent["type"] = key
                without_begin_end = deepcopy(next_ent)
                del without_begin_end["begin"]
                del without_begin_end["end"]
                yield without_begin_end, "container_ent"
                break
            elif next_ent["begin"] >= end:
                # span is after
                break
            elif next_ent["begin"] >= begin and next_ent["end"] <= end:
                # span is sub span of this span
                next_ent = values.pop()
                next_ent["begin_in_doc"] = next_ent["begin"]
                next_ent["end_in_doc"] = next_ent["end"]
                next_ent["begin"] -= begin
                next_ent["end"] -= begin
                next_ent["type"] = key
                yield next_ent, "sub_ent"
            else:
                # no container and no subspan => span has an overlap!
                if end >= next_ent["end"]:
                    values.pop()
                yield next_ent, "overlapping_ent"
                break


line_pattern = re.compile("[^\n]*[\n$]")


def generate_line_annos(text):
    lines = []
    for idx, line in enumerate(line_pattern.finditer(text)):
        if line.group().strip():
            lines.append(dict(begin=line.start(), end=line.end(), idx=idx))
    return lines



import json
from lxml import etree
from glob import glob
from pathlib import Path
from typing import Union, Optional, List, Dict

ns = dict(namespaces={"tei": "http://www.tei-c.org/ns/1.0"})
tei_prefix = "{http://www.tei-c.org/ns/1.0}"
xml_id = '{http://www.w3.org/XML/1998/namespace}id'

def transform_xml_folder(source_path: Path, target_path: Path, sentences=True) -> str:
    target_path_txt = target_path / Path("text_extracted")
    target_path_json = target_path / Path("text_extracted_json")
    target_path_txt.mkdir(parents=True, exist_ok=True)
    target_path_json.mkdir(parents=True, exist_ok=True)
    for file in glob(str(source_path) + "/*.xml"):
        try:
            text, annos = transform_xml(file, sentences)
            doc = dict(text=text, annotations=annos)
            fn = Path(file).stem
            fn = fn[:-4] if fn.endswith(".tei") else fn
            fn_txt = target_path_txt / Path(fn + ".txt")
            fn_json = target_path_json / Path(fn + ".json")
            fn_txt.open("w").write(text)
            json.dump(doc, fn_json.open("w"))
        except Exception as e:
            print("xml to text not working")
            print(file)
            print(e)

def transform_xml(filename, sentences=True):
    root = etree.parse(filename)
    extractor = TeiExtractor(root, sentences)
    doc, annotations = extractor.extract()
    return doc, annotations

class TeiExtractor:
    def __init__(self, root, sentences=True):
        self.text = ""
        self.annos = {}
        self.root = root
        self.sentences = sentences
       
    def extract(self):
        # reset current extraction
        self.text, self.annos = "", {}
        # header section
        self.add_extraction(self.get_title(), sep="")
        # doc = add_extraction(doc, get_authors(root), sep="")
        self.text += "\n"
        self.add_extraction(self.get_abstract(), sep="\n")
        self.text += "\n\nMain:\n"
        self.annos["SectionHeader"] = [dict(begin=0, end=len(self.text))]
        begin_main = len(self.text)
        self.add_extraction(self.get_body_divs(), sep="\n")
        self.annos["SectionMain"] = [dict(begin=begin_main, end=len(self.text))]
        self.text += "\n"
        self.add_extraction(self.get_footnotes(), sep="\n")
        self.text += "\n"
        self.add_extraction(self.get_references(), sep="\n")
        # add idx to each element
        for key_annos in self.annos.values():
            key_annos.sort(key=lambda x: (x["begin"], x["end"]))
            for idx, anno in enumerate(key_annos):
                anno["idx"] = idx
        return self.text, self.annos

    def add_extraction(self, new, sep=""):
        text, annos = _add_extraction(
            (self.text, self.annos),
            new, sep)
        self.text, self.annos = text, annos
        
    def get_title(self):
        xpath = "/tei:TEI/tei:teiHeader/tei:fileDesc/tei:titleStmt/tei:title"
        text, annos = "", {}
        title_ele = self.root.xpath(xpath, **ns)
        if title_ele:
            title, annos = self.get_text_with_refs(title_ele[0])
        return title, annos

    def get_abstract(self):
        xpath = "/tei:TEI/tei:teiHeader/tei:profileDesc/tei:abstract/tei:div"
        divs = self.root.xpath(xpath, **ns)
        text, annos = "Abstract:\n", {}
        #print("abstract")
        for idx, div in enumerate(divs):
            #print("div in abstract")
            div_content = self.get_div(div)
            #print("div in abstract after get_div")
            sep = "\n" if idx > 0 else ""
            text, annos = _add_extraction((text, annos), div_content, sep)
        annos["Abstract"] = [dict(begin=0, end=len(text))]
        #print("abstract finished")
        return text, annos

    def get_body_divs(self):
        xpath = "/tei:TEI/tei:text/tei:body/tei:div" # AAARRRGGG
        xpath = "/tei:TEI/tei:text/tei:body/tei:div"#text/body/tei:div"
        div_elements = self.root.xpath(xpath, **ns)
        content = ("", {})
        for div_element in div_elements:
            div_extraction = self.get_div(div_element)
            content = _add_extraction(content, div_extraction, "\n\n")
        return content

    def get_div(self, div_element):
        #print("get_div")
        text, annos = "", {}
        first_p = True
        for element in div_element:
            tag = element.tag.replace(tei_prefix, "")
            if tag == "head":
                begin = len(text)
                n = element.attrib.get("n")
                if n:
                    text += n + " "
                text += element.text 
                head_anno = dict(begin=begin, end=len(text))
                if n:
                    head_anno["n"] = n
                annos["Head"] = annos.get("Head", []) + [head_anno]
                text += "\n"
            elif tag == "formula":
                text += element.text
            elif tag == "p":
                #print("found p in div")
                paragraph = self.get_paragraph(element)
                content = (text, annos)
                sep = "" if first_p else "\n"
                text, annos = _add_extraction(content, paragraph, sep)
                first_p = False
            else:
                print("not handled:", tag)
                print(element.text)
                print(element.attrib)
                for child in element:
                    print(child.text)
                pass
                #print(f"\r{element.tag} is not handled", end="")
        annos["Div"] = [dict(begin=0, end=len(text))]
        return text, annos

    def get_paragraph(self, element):
        #print("found para", element.tag)
        if not self.sentences:
            text, annos = self.get_text_with_refs(element)
        else:
            text, annos = self.get_sentences(element)
        #print(f"|{text}|")    
        annos["Paragraph"] = [dict(begin=0, end=len(text))]
        return text, annos

    def get_sentences(self, element):
        text, annos = "", {}
        for idx, sent in enumerate(element):
            if idx != 0:
                text += " "
            sent_text, sent_annos = self.get_text_with_refs(sent)
            sent_annos["Sentence"] = [dict(begin=0, end=len(sent_text))]
            if sent_text:
                text, annos = _add_extraction((text, annos),
                                                   (sent_text, sent_annos))
        return text, annos


    def get_footnotes(self):
        text, annos = "Footnotes:", {}
        footnotes = self.root.xpath("//tei:text/tei:body/tei:note", **ns)
        for footnote in footnotes:
            text += "\n"
            ident = footnote.attrib.get(xml_id)
            n = footnote.attrib.get('n')
            if n:
                footnote_start = (f"{n}: ", {})
            else:
                footnote_start = ("", {})
            footnote_text = self.get_text_with_refs(footnote)
            footnote_text, footnote_annos = _add_extraction(footnote_start, footnote_text)
            footnote_anno = dict(begin=0, end=len(footnote_text))
            if ident:
                footnote_anno["id"] = ident
            if n:
                footnote_anno["n"] = n
            footnote_annos["Footnote"] = [footnote_anno]
            text, annos = _add_extraction((text, annos), (footnote_text, footnote_annos))
        annos["SectionFootnote"] = [dict(begin=0, end=len(text))]
        return text, annos


    def get_references(self):
        # div[type="references"] ?
        xpath = "/tei:TEI/tei:text/tei:back/tei:div/tei:listBibl/tei:biblStruct"
        references = self.root.xpath(xpath, **ns)
        text = "References:\n\n"
        annos = {}
        for idx, ref in enumerate(references):
            ident = ref.attrib.get(xml_id)
            notes = ref.xpath("./tei:note[@type='raw_reference']", **ns)
            ref_text = ""
            for note in notes:
                if note.text:
                    text += "- "
                    ref_text += note.text.strip()
            if ref_text or ident:
                ref_anno = dict(begin=0, end=len(ref_text))
                if ident:
                    ref_anno["id"] = ident
                ref_annos = dict(ReferenceString=[ref_anno])
                text, annos = _add_extraction((text, annos), (ref_text, ref_annos), "")
                if idx > 0:
                    text += "\n\n"
        annos["SectionReference"] = [dict(begin=0, end=len(text))]
        return text, annos

    def get_text_with_refs(self, element):
        #print(f"get Text from {element.tag}")
        text, annos = "", {}
        if element.text and element.text.strip():
            text += element.text.strip()
        with_ref = False
        first_p = True
        for child in element:
            tag = child.tag.replace(tei_prefix, "")
            if tag == "ref":
                text_ref = child.text.strip() if child.text is not None else ""
                ref_type = child.attrib.get("type")
                target = child.attrib.get("target")
                if ref_type == "bibr":
                    ref_type = "ReferenceToBib"
                elif ref_type == "foot":
                    ref_type = "ReferenceToFootnote"
                elif ref_type == "figure":
                    ref_type = "ReferenceToFigure"
                elif ref_type == "table":
                    ref_type = "ReferenceToTable"
                elif ref_type == "formula":
                    ref_type = "ReferenceToFormula"
                else:
                    print(ref_type + " not known")
                    ref_type = "ReferenceUnknown"
                child_anno = dict(begin=0, end=len(text_ref))
                if target:
                    child_anno["target"] = target
                child_annos = {ref_type: [child_anno]}
                if text_ref and text_ref[0] not in ",;.!?)]":
                    text += " "
                text, annos = _add_extraction((text, annos),
                                                   (text_ref, child_annos))
                if child.tail and child.tail.strip():
                    tail_text = child.tail.strip()
                    if tail_text[0] not in ",;.!?)]":
                        text += " "
                    text += child.tail.strip()
            elif tag == "p":
                paragraph = self.get_paragraph(child)
                content = (text, annos)
                sep = "" if first_p else "\n"
                text, annos = _add_extraction(content, paragraph, sep)
                first_p = False
            else:
                print(tag, "not known here")
                print()
                print(element)
                print(element.attrib)
                print("text", child.text)
                print(child.attrib)
                for c in child:
                    print(c.attrib)
                    print(c.text)
                print(child)
                print(self.root.getpath(child))

                print(tag, "not known here")
            #print(etree.tostring(child))
            #text = text[:text.rfind("\n")]
            #if child.text:
            #    text += child.text[:child.text.rfind("\n") if "\n" in child.text else None]
            #if child.tail is not None:
            #    text += child.tail[:child.tail.rfind("\n") if "\n" in child.tail else None]
            #with_ref = True
        #print(text)
        return text, annos
    
def _add_extraction(old, new, sep=""):
    offset = 0
    text, annos = old
    new_text, new_annos = new
    if new_text:
        text += sep 
        offset = len(text)
        text += new_text
    annos_updated = {}
    keys = set(annos) | set(new_annos)
    for key in keys:
        annos_key_old = annos.get(key, [])
        annos_key_new = new_annos.get(key, [])
        [a.update(dict(
            begin=a["begin"] + offset,
            end=a["end"] + offset)
                 )
         for a in annos_key_new]
        annos_key_all = annos_key_old + annos_key_new
        annos_key_all.sort(key=lambda x: (x["begin"], ["x.end"]))
        annos_updated[key] = annos_key_all
    return text, annos_updated




import json

from pathlib import Path
from .extract_parts import generate_line_annos

class FlatTeiSeparateLoader:
    def __init__(self, txt_path, anno_path):
        self.txt_path = txt_path
        self.anno_path = anno_path

    def load_flat_tei(self, document_name):
        annos = []
        text  = ""
        path_text = Path(self.txt_path) / Path(document_name)
        path_anno = Path(self.anno_path) / Path(document_name).with_suffix(".json")
        text = path_text.open().read()
        annos = json.load(path_anno.open())
        return dict(text=text, annotations=annos)

class FlatTeiLoader:
    def __init__(self, pub_path):
        self.pub_path = Path(pub_path)

    def load_flat_tei(self, document_name):
        """ 
        parameters: document_name identifier of document 
        """
        file_path = self.pub_path / Path(document_name).with_suffix(".json")
        doc = json.load(file_path.open())
        doc["annotations"]["Line"] = generate_line_annos(doc["text"])
        return doc




from setuptools import setup

setup(
    name='flattentei',
    version='0.1.0',    
    description='Transform tei xml to a simple standoff format',
    url='https://',
    author='Wolf Otto',
    author_email='wolfgang.otto@gesis.org',
    license='BSD 2-clause',
    packages=['flattentei'],
    install_requires=[
        "lxml>=4.9.1",
    ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
)
